//! Getting started with the ber crate.
//!
//! This module provides a short introduction to ASN.1 and the Basic Encoding
//! Rules and how the crate can be used to decode and encode values.
//! It is only for documentation and does not contain any code.
//!
//! The guide consists of the following sections:

pub mod asn1_and_ber;
pub mod decode;
pub mod encode;
